// tòmò web analytics
// Copyright since 2011 by Yves Goergen
// Website: http://unclassified.de/go/tomo
// Licence: GNU General Public License (GPL) Version 3
// Encoding: UTF-8 (without BOM) (auto-detect: °°°°°)
//
// Few small parts of this file were inspired by Piwik, the open-source web analytics software.
// Piwik is released under GPL license at http://piwik.org
//
// This file contains the browser JavaScript code that collects all client-side data and events
// and sends them to the tracking server script.

var tomo = function()
{
	// ===================================  CONFIGURATION DATA  ===================================


	// ID attribute of the script element loading this script file.
	var cfg_scriptId = "tomoScript";
	var cfg_historyInputId = "tomoHistoryInput";
	// Global variable name to read the page load start time from.
	var cfg_startTime = "_tomoStartTime";
	// Name of the cookie used to cookies support feature test. Only use regexp-safe characters.
	var cfg_testCookieName = "_tomotest";
	// Time to wait after an activity event until the user is considered idle, when the window/page
	// is considered focused, in milliseconds
	var cfg_activityTimeoutFocused = 2 * 60 * 1000;
	// Time to wait after an activity event until the user is considered idle, when the window/page
	// is NOT considered focused, in milliseconds
	var cfg_activityTimeoutUnfocused = 30 * 1000;
	
	// Enables debugging features. Saves space when disabled.
	// 0: disabled
	// 1: normal debugging data
	// 2: (reserved)
	// 3: extended event logging
	// 4: with mousemove event logging
	/** @const */
	var cfg_debug = 1;


	// =============================  VARIABLES AND DATA DEFINITIONS  =============================


	// NOTE: New state variables might need to be re-initialised in resetVars() for when the page
	//       is revisited from bfcache and all of its previous state values are still set, or when
	//       a new pageview is triggered through an API call.

	// Internal data:
	
	// Absolute URL to the tomo server. This will be computed.
	var serverUrl;
	// Indicates whether the load event has already occured
	var didLoad = false;
	// Indicates whether the DOMContentLoaded event has already occured
	var didDomLoad = false;
	// Indicates whether the page was initially prerendered. This state needs to be remembered to
	// know whether to tell the server to clear the prerender flag once the page has become
	// visible.
	var prerendered = false;
	// Indicates whether the user activity is considered started due to a relevant event
	var didStartActivity = false;
	// Indicates whether the unload event has already occured
	var didUnload = false;
	
	// Page info:
	
	// Collected page data
	var pageData = {};
	// Time it took to load the page, in milliseconds. This is calculated in the load event.
	var pageLoadTime;
	// Domain of the current page, used to determine outlinks
	var myDomain;
	// New URL to redirect to in order to clear campaign parameter from URL
	var clearCampaignUrl;
	// Displayed language of the page. May be detected from translation service websites, or set
	// with the setPageLanguage API function.
	var pageLanguage;
	// CSS classes to ignore as link click
	var ignoredLinkClasses = [];
	
	// Server reply data:
	
	// Pageview ID, generated by the server for the page view, returned and saved for
	// further reference on page actions (outlink, unload, ...)
	var pageviewId;
	
	// User activity data:
	
	// Time of the initialisation event, if not prerendering. Used in onSomethingMoved().
	var initTime;
	// Time of the last recognised user activity
	var lastActivityTime;
	// Time when the user activity has last started. When this is set, the total user activity
	// timespan on this page is the sum of activityTimespan and the time that has passed since
	// activityBaseTime. When this is unset, the value in activityTimespan is already current.
	var activityBaseTime;
	// Summed up time that the user was considered interacting with the page, in milliseconds
	var activityTimespan = 0;
	// Interval to check user activity, in milliseconds
	var activityInterval = 1000;
	// Did we already report starting inactivity to the server?
	var inactiveReported = false;
	// Last reported activityTimespan value
	var lastPingTime = 0;
	// Time of the last action report that potentially leaves the page. Don't send unload events
	// too shortly after such an action as it is likely a duplicate, only delaying things.
	var lastLeavingActionTime = 0;
	// Indicates whether the browser window/page is currently focused/active. This may not be very
	// reliable but we'll do our best. Don't rely on that value!
	var isWindowFocused = true;
	// Indicates whether the focusin/focusout events have been observed. If these events are
	// triggered by a browser, they must be used in spite of focus/blur events (Internet Explorer).
	var seenFocusInEvents = false;
	// Remembers timestamps for clicks on outlinks to detect double-clicks
	var outlinkTimes = {};
	
	// Custom variables data:
	
	var cvSendTimeout;
	
	// Shortened data keys (must be synchronised with the server-side script)
	var key_pageviewId        = 'pv';
	var key_visitorId         = 'vi';
	var key_newVisitor        = 'nv';
	var key_newSession        = 'ns';
	var key_pageTitle         = 'pt';
	var key_pageLanguage      = 'la';
	var key_utcOffset         = 'ut';
	var key_screenWidth       = 'sw';
	var key_screenHeight      = 'sh';
	var key_screenAvailWidth  = 'aw';
	var key_screenAvailHeight = 'ah';
	var key_screenColors      = 'sc';
	var key_cookies           = 'co';
	var key_pageLoadTime      = 'pl';
	var key_campaign          = 'ca';
	var key_pageUrl           = 'pu';
	var key_referrer          = 're';
	var key_viewportWidth     = 'vw';
	var key_viewportHeight    = 'vh';
	var key_flash             = 'fl';
	var key_java              = 'ja';
	var key_quicktime         = 'qt';
	var key_silverlight       = 'sl';
	var key_windowsmedia      = 'wm';
	var key_adblock           = 'ab';
	var key_prerender         = 'pr';
	var key_origPageviewId    = 'op';
	var key_activityTimespan  = 'at';
	var key_duplicatePageview = 'dp';
	var key_outlinkUrl        = 'ou';
	var key_downloadUrl       = 'du';
	var key_imgZoomUrl        = 'iu';
	var key_sectionExpand     = 'se';
	var key_siteSearch        = 'ss';

	// Shortable refererences to common objects
	var _document = document;
	var _navigator = navigator;
	var _screen = screen;
	var _window = window;
	var _location = location;


	// ====================================  HELPER FUNCTIONS  ====================================


	// Resets all variables to start a new pageview.
	//
	function resetVars()
	{
		initTime = undefined;
		_window[cfg_startTime] = undefined;
		pageLoadTime = undefined;
		didStartActivity = false;
		activityTimespan = 0;
		lastActivityTime = undefined;
		activityBaseTime = undefined;
		inactiveReported = false;
		lastPingTime = 0;
		lastLeavingActionTime = 0;
		pageviewId = undefined;   // Track a new pageview
		if (isSet(pageLanguage))
		{
			// Resend language that may have been previously set by a function call
			pageData[key_pageLanguage] = pageLanguage;
		}
	}
	
	// Determines whether a variable is set to a value.
	//
	// Returns (bool) true, if the value is set, false for undefined or null values.
	//
	function isSet(v)
	{
		return typeof v != "undefined" && v != null;
	}

	// Returns the Internet Explorer version, or 0 for other browsers.
	//
	// Source: http://msdn.microsoft.com/en-us/library/ms537509%28v=vs.85%29.aspx#4
	//
	function ieVersion()
	{
		var style = _document.documentElement.style;
		if (style.scrollbar3dLightColor != undefined)
		{
			if (style.opacity != undefined)
				return 9;
			else if (style.msBlockProgression != undefined)
				return 8;
			else if (style.msInterpolationMode != undefined)
				return 7;
			else if (style.textOverflow != undefined)
				return 6;
			else
				return 5.5;
		}
		return 0;
	}

	// Finds the event object in the DOM quirks.
	//
	// Returns the event object.
	//
	function findEvent(e)
	{
		e = e || event;   // IE
		if (!e.target && e.srcElement)   // IE
			e.target = e.srcElement;
		if (isSet(e.button))
		{
			// Every browser uses different values for the mouse buttons. Correct them here.
			// DOM says: 0 = left, 1 = middle, 2 = right (multiple buttons not supported)
			// Opera 7 and older and Konqueror are not specifically handled here.
			// See http://de.selfhtml.org/javascript/objekte/event.htm#button
			if (ieVersion() > 0)
			{
				if (e.button & 1)
					e.mouseButton = 0;
				else if (e.button & 2)
					e.mouseButton = 2;
				else if (e.button & 4)
					e.mouseButton = 1;
			}
			else
				e.mouseButton = e.button;
		}
		return e;
	}

	// Attaches an event to an element.
	//
	// The event handler will be wrapped in a compatibility layer offering a
	// DOM-compatible event object as the handler's parameter. If the handler
	// returns false, the event will be cancelled.
	//
	// The event handler function will be passed the event object as only argument.
	// The source element can be retrieved through e.target.
	//
	// element = (Element) Element to attach the event to
	// name = (string) Event name (DOM syntax, i.e. without "on")
	// handler = (function) Event handler function
	//
	// Returns a handle to allow detaching the event again.
	//
	function addEvent(element, name, handler)
	{
		// Page caching events are not supported everywhere, use old fallback events.
		if (name == "pageshow" && !("on" + name in _window))
		{
			// NOTE: Firefox (1.5 to 8) don't admit their support, so bfcache is (was...) not used
			// there.
			name = "load";
		}
		if (name == "pagehide" && !("on" + name in _window))
		{
			name = "unload";
		}

		// Vendor prefixes for the world... find the correct one.
		if (name == "visibilitychange" && !isSet(_document.hidden))
		{
			if (isSet(_document.mozHidden))
			{
				name = "moz" + name;
			}
			else if (isSet(_document.webkitHidden))
			{
				name = "webkit" + name;
			}
			else if (isSet(_document.msHidden))
			{
				name = "ms" + name;
			}
		}

		var wrappedHandler = function(e)
		{
			// Unify event object for DOM/IE
			e = findEvent(e);
			var res;
			try
			{
				res = handler(e);
				if (res === false)
				{
					// Cancel event bubbling - "don't tell my parents!"
					e.cancelBubble = true;   // IE
					if (e.stopPropagation)
						e.stopPropagation();   // DOM2

					// Prevent default action (doesn't work in Opera)
					e.returnValue = false;
					//try { e.preventDefault(); } catch (ex) { }   // DOM2
					if (e.preventDefault)
						e.preventDefault();   // TODO: Test
					return false;   // IE
				}
			}
			catch (ex)
			{
			}
			return true;   // IE
		};

		// Find a compatible event name for some less competent browsers
		var compatName = name;
		if (name == "DOMContentLoaded")
		{
			compatName = "load";
		}

		if (element.addEventListener)
			element.addEventListener(name, wrappedHandler, false);   // DOM, IE9
		else if (element.attachEvent)
			element.attachEvent("on" + compatName, wrappedHandler);   // IE
		else
			element["on" + compatName] = wrappedHandler;   // even older - Attention: This removes previously assigned events (?)

		return {"e": element, "n": name, "cn": compatName, "f": wrappedHandler};
	}

	// Removes an event from an element.
	//
	// h = Return value from previous addEvent call
	//
	function removeEvent(h)
	{
		if (h.e.addEventListener)
			h.e.removeEventListener(h.n, h.f, false);   // DOM
		else if (h.e.attachEvent)
			h.e.detachEvent("on" + h.cn, h.f);   // IE
		else
			h.e["on" + h.cn] = undefined;   // even older - Attention: This can only remove all events (?)
	}

	// Extracts a query string parameter from a URL.
	//
	// url = (string) URL to extract the parameter from.
	// name = (string) Parameter name.
	//
	function extractParameter(url, name)
	{
		var matches = new RegExp("^[fht]+ps?:/*[^?]+[?]([^#]+)").exec(url);
		var result = matches ? new RegExp("(?:^|&)" + name + "=([^&]*)").exec(matches[1]) : undefined;
		return isSet(result) ? decodeURIComponent(result[1]) : undefined;
	}

	// Extracts the host name from a URL.
	//
	// url = (string) URL to extract the hostname from.
	// nofallback = (bool) Set true to not return the full URL if it cannot be parsed.
	//
	function extractHostName(url, nofallback)
	{
		var matches = new RegExp("^[fht]+ps?:/*(?:[^@]+@)?([^:/?#]+)").exec(url);
		return matches ? matches[1] : (nofallback ? '' : url);
	}

	// Gets the current visibility state of the page.
	//
	// This function handles different vendor prefixes automatically.
	//
	// Returns "prerender", "hidden" or "visible" (default if API is unsupported).
	//
	function getVisibilityState()
	{
		if (isSet(_document.hidden))
		{
			return _document.visibilityState;
		}
		else if (isSet(_document.mozHidden))
		{
			return _document.mozVisibilityState;
		}
		else if (isSet(_document.webkitHidden))
		{
			return _document.webkitVisibilityState;
		}
		else if (isSet(_document.msHidden))
		{
			return _document.msVisibilityState;
		}
		return "visible";
	}
	
	// Updates the user activity base time and applies the elapsed time accordingly.
	//
	function setActivity(active)
	{
		var now = (new Date()).getTime();
		if (active)
		{
			didStartActivity = true;
			lastActivityTime = now;
			if (!isSet(activityBaseTime))
			{
				// Only set the new base time if it is currently unset.
				activityBaseTime = now;
			}
		}
		else
		{
			lastActivityTime = 0;
			if (isSet(activityBaseTime))
			{
				// The activity base time is currently set and the user is now considered inactive.
				// Apply the elapsed timespan and unset the base time.
				activityTimespan += now - activityBaseTime;
				activityBaseTime = undefined;
			}
		}
	}
	
	// Gets the current user activity timespan, which may need to be computed since
	// activityBaseTime.
	//
	function getActivityTimespan()
	{
		var at = activityTimespan;
		if (isSet(activityBaseTime))
		{
			at += (new Date()).getTime() - activityBaseTime;
		}
		return at;
	}
	
	// Schedules sending of data for custom variables in a short time, in case no ping is already
	// sent.
	//
	function scheduleCVSend()
	{
		if (isSet(cvSendTimeout))
		{
			clearTimeout(cvSendTimeout);
		}
		cvSendTimeout = setTimeout(postData, 500);
	}

	// Writes a message to the log data field. Its time is recorded and all written messages are
	// sent to the server with the next data posting.
	//
	function writeLog(msg)
	{
		var now = new Date();
		var ms = now.getMilliseconds();
		var nowStr = now.toLocaleTimeString() + ".";
		if (ms < 10)
			nowStr += "00" + ms;
		else if (ms < 100)
			nowStr += "0" + ms;
		else
			nowStr += ms;
		
		if (isSet(pageData.log))
		{
			pageData.log += "\n";
		}
		else
		{
			pageData.log = "";
		}
		pageData.log += nowStr + " " + msg;
		scheduleCVSend();
	}


	// =================================  COMMUNICATION FUNCTIONS  ================================


	// Sends an asynchronous HTTP POST request.
	//
	// url = (string) URL to send the request to.
	// data = (object) Associative array of the data to send, only scalars or flat arrays allowed as values.
	// handler = (function) Return event handler.
	// async = (bool) Asynchronous request (optional, default: true)
	//
	// Returns (boolean) true if the request was sent, false otherwise.
	//
	function postRequest(url, data, handler, async)
	{
		// Set default parameters
		if (!isSet(async)) async = true;

		// Assemble the data values to form submit format
		var dataToSend = "";
		for (var property in data)
		{
			var item = data[property];
			if (typeof item == "number" || typeof item == "string")
			{
				dataToSend += (dataToSend != "" ? "&" : "") +
					property + "=" + encodeURIComponent(item);
			}
		}

		function imageRequest(reason)
		{
			if (isSet(reason))
			{
				// trim function compatibility replacement (from Mozilla Developer site)
				reason = "&reason=" + encodeURIComponent(reason.replace(/^\s+|\s+$/g, ""));
			}

			var image = new Image(1, 1);
			addEvent(image, "load", function() { if (handler) handler(); });
			image.src = url + (url.indexOf("?") >= 0 ? "&" : "?") + dataToSend + "&img=1" + reason + "&rnd=" + Math.round(Math.random() * 999999);

			/*if (!async)
			{
				// Block UI for a short time so that the last HTTP request still goes out
				var now = new Date();
				var expireDateTime = now.getTime() + 100;   // milliseconds
				do
				{
					now = new Date();
				}
				while (now.getTime() < expireDateTime);

				// Or try the following code:
				// Source: http://forums.thedailywtf.com/forums/t/24374.aspx
				var s = e = new Date();
				for (var i = 0; e.getTime() - s.getTime() < 250; i++)
				{
					e = new Date();
				}
				// Another solution: http://dev.piwik.org/trac/ticket/2929
			}*/
		}

		// Get an XmlHttpRequest instance
		var request;
		var needPlainTextMarker = false;
		var needCrossDomain = extractHostName(url) != extractHostName(_location.href);

		if (_window.XDomainRequest && needCrossDomain)   // IE8+, allows cross-domain requests
		{
			request = new _window.XDomainRequest();
			//needPlainTextMarker = true;
		}
		else if (_window.XMLHttpRequest)   // DOM, allows cross-domain requests in Firefox 3.5+, Safari4+
		{
			request = new _window.XMLHttpRequest();
		}
		else if (_window.ActiveXObject)   // older IE
		{
			try
			{
				request = new ActiveXObject("Msxml2.XMLHTTP");   // IE 6 (and others?)
			}
			catch (ex)
			{
				try
				{
					request = new ActiveXObject("Microsoft.XMLHTTP");
				}
				catch (ex2)
				{
				}
			}
		}

		if (!request)
		{
			// Browser does not support any kind of XMLHttpRequest object
			imageRequest(cfg_debug >= 1 ? "no XMLHttpRequest support" : "");
			return false;
		}

		// Submit data
		try
		{
			request.open("POST", url, async);
			request.onreadystatechange = function()
			{
				if (request.readyState == 4)
				{
					if (request.status == 200)
					{
						// Do something with the response
						try
						{
							if (handler) handler(request);
						}
						catch (ex)
						{
							console.log("tomo: postRequest: error: " + ex);
						}
					}
					else
					{
						// Server responded with error code request.status (request.statusText)
						//alert("ERROR: Server responded with error code " + request.status + " (" + request.statusText + ")");
						// Try another way
						imageRequest(cfg_debug >= 1 ? "code=" + request.status + ",text=" + request.statusText : "");
					}
				}
			};
			try
			{
				request.setRequestHeader('Content-Type', 'application/x-www-form-urlencoded');
			}
			catch (ex)
			{
				// This method is not supported by Microsoft's Msxml2.XMLHTTP and XDomainRequest
				needPlainTextMarker = true;
			}
			request.send((needPlainTextMarker ? "!POST&" : "") + dataToSend);
		}
		catch (ex)
		{
			// Something went wrong... Try another way
			imageRequest(cfg_debug >= 1 ? "postRequest exception: " + ex.message : "");
		}
		return true;
	}


	// Posts any pending data to the server.
	//
	// This function can be called in any situation to send any data yet unsent to the server.
	// It can send combined data of multiple events in a single request and keeps track about
	// what data still needs to be sent.
	//
	// leaving = (bool) true if the page is about to unload, so don't use async posts then.
	//
	function postData(leaving, handler, retryCount)
	{
		// Set default parameters
		if (!isSet(retryCount)) retryCount = 0;

		// If we're retrying to send the data (because the pageview ID response has still not
		// arrived) for more than a second, we'll better give up.
		if (retryCount > 10)
		{
			if (isSet(handler)) handler();
			return;
		}
		
		if (isSet(pageviewId))
		{
			// Always send the pageview ID with the request if we have one, that's what it's for.
			pageData[key_pageviewId] = pageviewId;
		}
		else if (!isSet(pageData[key_pageUrl]))
		{
			// The initial page data has already been sent but the pageview ID response has not yet
			// arrived. We have a little problem now because we cannot send that request right now.
			// We can only retry later.
			setTimeout(function() { postData(leaving, handler, retryCount + 1); }, 100);
			// Don't touch any data so that it's still there next time.
			return;
		}

		// Always include the current activity timespan with the data
		pageData[key_activityTimespan] = getActivityTimespan();
		// Remember when the last activity timespan report was. This is used to determine when the
		// next ping request should be sent, in case pinging is currently active (see other uses of
		// that variable).
		lastPingTime = pageData[key_activityTimespan];

		// Make sure we know where to post to, but doing that once is enough.
		if (!isSet(serverUrl))
		{
			serverUrl = _document.getElementById(cfg_scriptId).src;
			serverUrl += (serverUrl.indexOf("?") != -1 ? "&" : "?") + "post=1";
		}
		
		postRequest(serverUrl, pageData, function(request) {
			// Server response handler
			if (isSet(request) && request.responseText.length > 0)
			{
				var response = eval("(" + request.responseText + ")");
				if (response[key_pageviewId])
				{
					pageviewId = response[key_pageviewId];
					// Remember pageview ID for future navigation back to this page
					var inp = _document.getElementById(cfg_historyInputId);
					if (inp && inp.value === "")
					{
						inp.value = pageviewId;
					}
					if (response[key_activityTimespan])
					{
						activityTimespan += parseInt(response[key_activityTimespan]);
					}
				}
				if (response[key_newVisitor] && response[key_newSession])
				{
					// Send the first heartbeat in a short random time to possibly correct the cookie in the database
					// TODO: Maybe synchronise a bit with other events that send data? Like with Opera special activity notifications
					//setTimeout(function() { postData(); }, Math.round(1000 + Math.random() * 2000));
					// NOTE: Pinging is always enabled now, with the first ping after ~5 seconds
				}
			}
			
			// Allow the caller further response handling
			if (isSet(handler)) handler(response);
		}, !leaving);
		
		// Clear all collected data to not send it again
		pageData = {};

		// Cancel scheduled sending for custom variables, see scheduleCVSend()
		if (isSet(cvSendTimeout))
		{
			clearTimeout(cvSendTimeout);
		}
	}


	// =============================  DATA AND ENVIRONMENT FUNCTIONS  =============================


	function collectWindowSize(retryCount)
	{
		// Set default parameters
		if (!isSet(retryCount)) retryCount = 0;

		// If we're retrying to send the data (because the pageview ID response has still not
		// arrived) for more than a second, we'll better give up.
		if (retryCount > 10) return;
		
		pageData[key_viewportWidth] = _window.innerWidth;
		pageData[key_viewportHeight] = _window.innerHeight;
		// Internet Explorer special cases
		if (!pageData[key_viewportWidth] && !pageData[key_viewportHeight])
		{
			pageData[key_viewportWidth] = _document.documentElement.clientWidth;
			pageData[key_viewportHeight] = _document.documentElement.clientHeight;
			if (!pageData[key_viewportWidth] && !pageData[key_viewportHeight])
			{
				pageData[key_viewportWidth] = _document.body.clientWidth;
				pageData[key_viewportHeight] = _document.body.clientHeight;
			}
		}

		if (cfg_debug >= 1) pageData["dbg_framed"] = window != window.top ? 1 : 0;

		if (pageData[key_viewportWidth] < 50 || pageData[key_viewportHeight] < 50)
		{
			// Bing image search includes the actual page in an iframe. The iframe element needs
			// some time to unfold and reveal its actual size. It has also been observed that the
			// viewport size was 10x10 px which seems too small and could be expanded later.
			// If no reasonable size has been determined, retry in a while.
			setTimeout(function() { collectWindowSize(retryCount + 1); }, 1000);
		}
		else if (retryCount > 0)
		{
			// This needs to be sent separately. Schedule it along with custom variables to be sent
			// to the server.
			scheduleCVSend();
		}
	}
	
	// Collects all data related to the current page and saves it into pageData. This function is
	// called every time a page is initially loaded.
	//
	// This includes:
	// * Campaign
	// * Window size
	// * Screen size
	// * UTC offset
	// * Cookie support
	// * Plugins
	// * Ad blocker
	// * Prerender state
	//
	function collectFirstPageData()
	{
		// Recognise campaign in query parameter or hash
		//
		// If a campaign parameter was recognised, its value is stored in pageData and a new URL
		// without the campaign parameter is generated and written to clearCampaignUrl. This variable
		// may later be used to start a redirection.
		//
		var refParam = extractParameter(_location.href, "ref");
		if (isSet(refParam))
		{
			pageData[key_campaign] = refParam;
			// Remove ref query parameter from URL
			clearCampaignUrl = _location.href.replace(/\?.*$/, _location.search.replace(/(\?|&)ref=[^&]*/, "$1")).
				replace(/[?&]+$/, "").
				replace(/&&/, "&").
				replace(/\?&/, "?");
		}
		else
		{
			var result = new RegExp("(?:#|;)ref=([^;]*)").exec(_location.hash);
			if (result)
			{
				pageData[key_campaign] = decodeURIComponent(result[1]);
				// Remove everything from the ref to the end from the hash
				var newHash = _location.hash.replace(/(#|;)ref=[^;]*/, "$1").
					replace(/[#;]+$/, "").
					replace(/;;/, ";").
					replace(/#;/, "#");
				clearCampaignUrl = _location.href.replace(/#.*$/, newHash);
			}
		}
		
		// Window size
		collectWindowSize();

		// Screen size
		pageData[key_screenWidth] = _screen.width;
		pageData[key_screenHeight] = _screen.height;
		pageData[key_screenAvailWidth] = _screen.availWidth;
		pageData[key_screenAvailHeight] = _screen.availHeight;
		pageData[key_screenColors] = _screen.colorDepth;

		// UTC offset
		pageData[key_utcOffset] = -new Date().getTimezoneOffset();   // UTC offset in minutes

		// Cookie support
		// NOTE: navigator.cookieEnabled is not reliable because it does not respect site
		//       exceptions. So instead we first watch out if there are already cookies
		//       (in that case they must be supported), and if not, we set, read and delete a
		//       test cookie. We can only find out whether cookies are at least accepted for
		//       the session. It cannot be determined whether they will persist for the next
		//       session.
		if (_document.cookie != "")
		{
			pageData[key_cookies] = 1;
		}
		else
		{
			// Set test cookie
			_document.cookie = cfg_testCookieName + "=1";
			pageData[key_cookies] = _document.cookie.match(new RegExp("(^|;)\\s*" + cfg_testCookieName + "=1\\s*(;|$)")) ? 1 : 0;
			// Delete test cookie
			var expireDate = new Date(0);
			_document.cookie = cfg_testCookieName + "=1;expires=" + expireDate.toUTCString();
		}

		// Plugins
		pageData[key_flash] = 0;
		pageData[key_java] = _navigator.javaEnabled() ? 1 : 0;
		pageData[key_quicktime] = 0;
		pageData[key_silverlight] = 0;
		pageData[key_windowsmedia] = 0;

		for (var i = 0; i < _navigator.mimeTypes.length; i++)
		{
			var type = _navigator.mimeTypes[i].type;
			if (type.match(/^application\/x-shockwave-flash$/)) pageData[key_flash] = 1;
			if (type.match(/^video\/quicktime$/)) pageData[key_quicktime] = 1;
			if (type.match(/^application\/x-silverlight$/)) pageData[key_silverlight] = 1;
			if (type.match(/^application\/x-mplayer2$/)) pageData[key_windowsmedia] = 1;
		}

		// Internet Explorer needs some special care and only reveals its plugins if asking louder.
		// Source:
		// http://www.builtfromsource.com/2007/06/26/detecting-plugins-in-internet-explorer-and-a-few-hints-for-all-the-others/
		if (_window.ActiveXObject)
		{
			if (!pageData[key_flash])
			{
				try
				{
					var control = new ActiveXObject('ShockwaveFlash.ShockwaveFlash');
					pageData[key_flash] = isSet(control) ? 1 : 0;
				}
				catch (ex)
				{
				}
			}
			if (!pageData[key_quicktime])
			{
				try
				{
					var control = new ActiveXObject('QuickTime.QuickTime');
					pageData[key_quicktime] = isSet(control) ? 1 : 0;
				}
				catch (ex)
				{
				}
			}
			if (!pageData[key_silverlight])
			{
				try
				{
					var control = new ActiveXObject('AgControl.AgControl');
					pageData[key_silverlight] = isSet(control) ? 1 : 0;
				}
				catch (ex)
				{
				}
			}
			if (!pageData[key_windowsmedia])
			{
				try
				{
					var control = new ActiveXObject('WMPlayer.OCX');
					pageData[key_windowsmedia] = isSet(control) ? 1 : 0;
				}
				catch (ex)
				{
				}
			}
		}

		// Ad blocker
		var ad = _document.createElement("div");
		ad.style.position = "absolute";
		ad.style.left = "0";
		ad.style.top = "-100px";
		ad.style.width = "20px";
		ad.setAttribute("class", "adbottom");
		_document.body.appendChild(ad);
		pageData[key_adblock] = ad.offsetWidth ? 0 : 1;
		_document.body.removeChild(ad);

		// Prerender state
		if (getVisibilityState() == "prerender")
		{
			pageData[key_prerender] = 1;
			prerendered = true;
		}
		if (cfg_debug >= 1 && getVisibilityState() == "hidden")
		{
			pageData["dbg_vis_hidden"] = 1;
		}
	}
	
	// Collects all data related to the current page and saves it into pageData. This function is
	// called whenever a page is shown from the bfcache, e.g. when navigating back in history. This
	// does not necessarily need to happen, depending on how the browser caches the previously
	// loaded page. When the user comes back to this page, a new pageview must be tracked, with the
	// data collected in this function. This is only a small subset of all data collected.
	//
	// This includes:
	// * Referrer
	// * Page title
	// * Page URL
	// * Navigation history
	//
	function collectPageData()
	{
		// Referrer URL
		pageData[key_referrer] = "";
		try
		{
			pageData[key_referrer] = _window.top.document.referrer;
		}
		catch (ex)
		{
			if (_window.parent)
			{
				try
				{
					pageData[key_referrer] = _window.parent.document.referrer;
				}
				catch (ex2)
				{
					pageData[key_referrer] = "";
				}
			}
		}
		if (pageData[key_referrer] === "")
		{
			pageData[key_referrer] = _document.referrer;
		}

		var orefParam = extractParameter(_location.href, "oref");
		if (isSet(orefParam))
		{
			// Manual old-site redirection. This parameter contains the original referrer if it is
			// still available. We notice this referrer instead of the URL of our own old site
			// which is redirecting here. The campaign redirect mechanism is used to get rid of the
			// URL parameter. The old site's domain is set as campaign name.
			pageData[key_campaign] = extractHostName(pageData[key_referrer]);   //"oldsite-redirect";
			pageData[key_referrer] = orefParam;
			// Remove oref query parameter from URL
			clearCampaignUrl = _location.href.replace(/\?.*$/, _location.search.replace(/(\?|&)oref=[^&]*/, "$1")).
				replace(/[?&]+$/, "").
				replace(/&&/, "&").
				replace(/\?&/, "?");
		}
		var rrefParam = extractParameter(_location.href, "rref");
		if (isSet(rrefParam))
		{
			// Automatic old-site redirection. This parameter contains the redirecting referrer if
			// it is available. The campaign redirect mechanism is used to get rid of the URL
			// parameter. This is the old site's URL that redirected here and its domain is set as
			// campaign name. The original referrer is still available from the HTTP header in this
			// case of an automatic redirection.
			pageData[key_campaign] = extractHostName(rrefParam);   //"oldsite-redirect";
			// Remove oref query parameter from URL
			clearCampaignUrl = _location.href.replace(/\?.*$/, _location.search.replace(/(\?|&)rref=[^&]*/, "$1")).
				replace(/[?&]+$/, "").
				replace(/&&/, "&").
				replace(/\?&/, "?");
		}

		// Page title
		pageData[key_pageTitle] = _document.title;
		
		// Page URL
		pageData[key_pageUrl] = _location.href;
		
		// Remember my own domain, we need it to recognise outlinks
		myDomain = _document.domain.replace(/\.+$/, "");   // Cut "." from the end
		// NOTE: We don't set myDomain because when translating the page, internal links are
		//       effectively also translated to the domain of the translation service website.
		//       If we fixed the value of myDomain here, all internal links would be detected as
		//       outlinks, logging a separate link click action.
		if (myDomain === "translate.googleusercontent.com")   // Google
		{
			if (pageData[key_referrer] === "")
			{
				pageData[key_referrer] = pageData[key_pageUrl];
			}
			pageLanguage = extractParameter(pageData[key_pageUrl], "tl");
			pageData[key_pageUrl] = extractParameter(pageData[key_pageUrl], "u");
			pageData[key_pageLanguage] = pageLanguage;
			//myDomain = extractHostName(pageData[key_pageUrl]);
		}
		// TODO: Add Bing Translator, but it's broken for XHTML pages, so cannot test right now.
		else if (myDomain === "www.babelfish.de")   // babelfish.de
		{
			if (pageData[key_referrer] === "")
			{
				pageData[key_referrer] = pageData[key_pageUrl];
			}
			pageLanguage = extractParameter(pageData[key_pageUrl], "dst");
			pageData[key_pageUrl] = extractParameter(pageData[key_pageUrl], "url");
			pageData[key_pageLanguage] = pageLanguage;
			//myDomain = extractHostName(pageData[key_pageUrl]);
		}
		else if (myDomain === "ceviri.yandex.net")   // Yandex (?)
		{
			if (!pageData[key_viewportWidth])
			{
				// Cancel tracking, this has been observed a duplicate request. That translation
				// service seems to severely interfer with the tracking process.
				pageData[key_pageUrl] = undefined;
			}
			else
			{
				if (pageData[key_referrer] === "")
				{
					pageData[key_referrer] = pageData[key_pageUrl];
				}
				var m = pageData[key_pageUrl].match(/^https?:\/\/[^\/]+\/tr-url\/..[+-](..)\.tr\/(.+)$/);
				if (m)
				{
					pageLanguage = m[1];
					pageData[key_pageUrl] = "http://" + m[2];
					pageData[key_pageLanguage] = pageLanguage;
				}
				//myDomain = extractHostName(pageData[key_pageUrl]);
			}
		}
		else if (myDomain === "cc.bingj.com" ||   // Bing
				myDomain === "webcache.googleusercontent.com" ||   // Google
				myDomain.substr(0, 5) === "74.6.")   // Yahoo (via Inktomi 74.6.0.0/16)
		{
			pageData[key_pageUrl] = _document.links[0].href;
			//myDomain = extractHostName(pageData[key_pageUrl]);
		}

		// Navigation history
		var inp = _document.getElementById(cfg_historyInputId);
		if (inp && inp.value !== "")
		{
			pageData[key_origPageviewId] = inp.value;
		}
	}


	// =================================  EVENT HANDLER FUNCTIONS  ================================


	// Page initialisiation "event", called when the tracking function has been called in the page.
	// This is during parsing the HTML tree, right after loading the tracker script.
	//
	function init()
	{
		// Collect all data. Everything should be available for us now.
		collectFirstPageData();
		collectPageData();

		if (!isSet(pageData[key_pageUrl])) return;   // Cancel tracking for this page
		
		if (isSet(clearCampaignUrl))
		{
			// This is a campaign URL. Send the data now, disable all other events and then start
			// the redirection to get to the real page.
			postData(false, function() {
				// Clear campaign parameter from URL, now that the page request has been tracked
				_location.replace(clearCampaignUrl);
			});
			return;
		}
		
		// Register all event handlers and start off the load timeout to make sure we send
		// _something_ even if loading takes forever.
		addEvent(_window, "DOMContentLoaded", onDOMContentLoaded);
		addEvent(_window, "pageshow", onPageshow);
		addEvent(_window, "focus", onFocus);
		addEvent(_window, "blur", onBlur);
		addEvent(_document, "focusin", onFocusin);
		addEvent(_document, "focusout", onFocusout);
		addEvent(_document, "visibilitychange", onVisibilitychange);
		addEvent(_window, "pagehide", onPagehide);
		setTimeout(onLoadTimeout, 2000);

		// Register user activity event handlers
		addEvent(_document, 'click', onSomethingMoved);
		addEvent(_document, 'mouseup', onSomethingMoved);
		addEvent(_document, 'mousedown', onSomethingMoved);
		addEvent(_document, 'mousemove', onSomethingMoved);
		addEvent(_document, 'mousewheel', onSomethingMoved);
		addEvent(_window, 'DOMMouseScroll', onSomethingMoved);
		addEvent(_document, 'wheel', onSomethingMoved);
		addEvent(_window, 'scroll', onSomethingMoved);
		addEvent(_document, 'keypress', onSomethingMoved);
		addEvent(_document, 'keydown', onSomethingMoved);
		addEvent(_document, 'keyup', onSomethingMoved);
		// TODO: check other complete implementation of user activity tracking
		//       http://stackoverflow.com/a/9502074/143684

		// Active activity polling:
		// This function checks every short period of time whether the last user activity was
		// within the activity grace period. Changes in that state are reported to the server.
		setInterval(onTick, activityInterval);

		// Remember when we started things, it may be useful later (in onSomethingMoved()).
		if (!prerendered)
		{
			initTime = (new Date()).getTime();
		}
	}
	
	// Event handler for the window.DOMContentLoaded or window.load event.
	//
	function onDOMContentLoaded(e)
	{
		if (didDomLoad)
		{
			if (cfg_debug >= 3) writeLog("onDOMContentLoaded, e.type=" + e.type + ", didDomLoad is already set")
			return;
		}
		didDomLoad = true;

		if (cfg_debug >= 3) writeLog("onDOMContentLoaded, e.type=" + e.type)

		// Add link click tracking to detect outlinks and downloads. This can be done as soon as
		// the DOM tree has loaded and doesn't have to wait for all page resources (images etc.)
		// are loaded (in the window.load event).
		if (_document.links)
		{
			for (var i = 0; i < _document.links.length; i++)
			{
				var link = _document.links[i];
				
				// Ignore CSS classes
				var ignoreLink = false;
				var linkClasses = link.className.split(" ");
				for (var j = 0; j < linkClasses.length && !ignoreLink; j++)
				{
					for (var k = 0; k < ignoredLinkClasses.length && !ignoreLink; k++)
					{
						if (linkClasses[j] == ignoredLinkClasses[k])
							ignoreLink = true;
					}
				}
				if (ignoreLink) continue;

				// Add all relevant events to a link to detect outgoing links or downloads
				addEvent(link, "click", onLinkClicked);
				//addEvent(link, "mousedown", onLinkClicked);
				// TODO: mousedown comes before click. A synchronous AJAX request makes Firefox miss the mouseup event
				//       so that it won't trigger the click event and eventually, no navigation is started. The user
				//       is annoyed and needs to click again. You need to click slowly or reclick because no second
				//       AJAX request is done in a short time.
				
				// TODO: Add some API to register new links to monitor that may be created at runtime
			}
		}
	}

	// Event handler for the window.pageshow or window.load event.
	//
	function onPageshow(e)
	{
		// In case we came back to a living page, try to clean up its state for another run.
		if (didUnload)
		{
			if (cfg_debug >= 3) writeLog("onPageshow, e.type=" + e.type + ", didUnload is set -> cleanup and restart")

			didUnload = false;
			didLoad = false;

			// Collect updated data to submit for a new pageview
			collectPageData();
			
			// Everything else that init() did for us is still there, so there's no need to perform
			// further initialisation after this cleanup.
			resetVars();
		}
		
		if (didLoad)
		{
			if (cfg_debug >= 3) writeLog("onPageshow, e.type=" + e.type + ", didLoad is already set")
			return;
		}
		didLoad = true;

		if (cfg_debug >= 3) writeLog("onPageshow, e.type=" + e.type)

		// The page has fully loaded. Measure the load time, then send the collected data to the
		// server, if it hasn't already been done in onLoadTimeout.

		// Measure load time, if possible
		if (isSet(_window[cfg_startTime]))
		{
			pageLoadTime = (new Date()).getTime() - _window[cfg_startTime].getTime();
			pageData[key_pageLoadTime] = pageLoadTime;
		}
		
		if (!prerendered || _window.opera)
		{
			// Some browsers (Chrome 21, Opera 12, Safari 5.1) don't trigger a window.focus event
			// if the page has loaded in the foreground. They did trigger that event though when
			// the page loaded in a background tab or in prerender and then is switched to.
			// (Safari is not covered by the above condition.)
			isWindowFocused = true;
		}
		
		// Send data
		postData();
	}
	
	// Called some time after the script has initialised. Ensures that something is done in a
	// reasonable time after page loading has started, even if there was no window.load event yet.
	//
	function onLoadTimeout()
	{
		if (didLoad)
		{
			if (cfg_debug >= 4) writeLog("onLoadTimeout, didLoad is already set")
			return;
		}

		if (cfg_debug >= 3) writeLog("onLoadTimeout")

		if (cfg_debug >= 1) pageData["dbg_lastEvent"] = "onLoadTimeout";

		// The page has no finished loading yet, so no pageview data was sent yet. Do that now.
		postData();
	}
	
	// Event handler for the window.focus event.
	//
	function onFocus(e)
	{
		if (cfg_debug >= 3) writeLog("onFocus, e.type=" + e.type + ", prerendered=" + (prerendered ? 1 : 0))

		// The window has focused. If it is the first time, the page is actually visible from now
		// on. Start user activity tracking. Don't do it if in prerender mode, though. The
		// corresponding visibilitychange event will be fired next and do the same.
		if (!prerendered)
		{
			setActivity(true);
			isWindowFocused = true;
		}
	}
	
	// Event handler for the window.blur event.
	//
	function onBlur(e)
	{
		if (!seenFocusInEvents)
		{
			isWindowFocused = false;
		}
	}
	
	// Event handler for the document.focusin event (Internet Explorer).
	//
	function onFocusin()
	{
		if (_window.opera) return;   // Don't fool us! You will never be like Internet Explorer.
		
		seenFocusInEvents = true;

		if (!prerendered)
		{
			setActivity(true);
			isWindowFocused = true;
		}
	}
	
	// Event handler for the document.focusout event (Internet Explorer).
	//
	function onFocusout()
	{
		if (_window.opera) return;   // Don't fool us! You will never be like Internet Explorer.
		
		seenFocusInEvents = true;
		isWindowFocused = false;
	}
	
	// Event handler for the document.visibilitychange event.
	//
	function onVisibilitychange()
	{
		if (cfg_debug >= 3) writeLog("onVisibilitychange, visibility=" + getVisibilityState())

		if (getVisibilityState() == "visible")
		{
			if (prerendered)
			{
				// Notify the server once that the prerendered page has become visible
				pageData[key_prerender] = 0;
				postData();
				prerendered = false;
				initTime = (new Date()).getTime();
			}

			// Start or resume user activity tracking
			setActivity(true);
			isWindowFocused = true;
		}
		else
		{
			// Pause user activity tracking
			setActivity(false);
		}
	}
	
	// Event handler for the window.pagehide or window.unload event.
	//
	function onPagehide(e)
	{
		if (didUnload)
		{
			if (cfg_debug >= 3) writeLog("onPagehide, e.type=" + e.type + ", didUnload is already set")
			return;
		}
		didUnload = true;
		
		if (cfg_debug >= 3) writeLog("onPagehide, e.type=" + e.type)

		var now = (new Date()).getTime();
		if (lastLeavingActionTime + 3000 > now)
		{
			// An action that has potentially left the page was reported within the last few
			// seconds. Don't repeat the unload event because it doesn't add much information and
			// in fact the user probably wasn't reading this page anyway after clicking the
			// outlink.
			return;
		}
		
		// The page is closing, be quick now. We need to send the latest user activity timespan to
		// to the server. Unfinished loading can also be detected now.
		if (!didLoad)
		{
			// Indicate unfinished loading by using a negative page load time value
			if (isSet(_window[cfg_startTime]))
			{
				pageLoadTime = now - _window[cfg_startTime].getTime();
				pageLoadTime = -pageLoadTime;
			}
			else
			{
				pageLoadTime = -1;
			}
			pageData[key_pageLoadTime] = pageLoadTime;
		}
		
		if (cfg_debug >= 1) pageData["dbg_lastEvent"] = "onPagehide";

		// Send data synchronously
		postData(true);
	}
	
	// Event handler for any registered user activity events.
	//
	function onSomethingMoved(e)
	{
		if (getVisibilityState() != "visible")
		{
			// Now that's really stupid. How can any input event occur when the page isn't even
			// visible?! Ignore it...
			return;
		}

		var now = new Date();

		if (cfg_debug >= 4 ||
			cfg_debug >= 3 && e.type != "mousemove")
			writeLog("onSomethingMoved, e.type=" + e.type)
		
		// If the page is visible and the activity is not considered started (because no focus or
		// visibilitychange event occured) before the first user activity is noticed, then the
		// relevant events were missing and thus any user presence (not necessarily activity!)
		// since displaying the page was not counted. To correct that the best we can, the time
		// between initialisation and this first user activity is added to the activity timespan.
		if (!didStartActivity && !lastActivityTime && isSet(initTime))
		{
			activityTimespan += now.getTime() - initTime;
			if (cfg_debug >= 3) writeLog("Added pre-load init time to activity timespan")
		}

		// The user has shown some activity. Start or resume user activity tracking.
		setActivity(true);
	}
	
	// Page tick handler. Called regularly to update the user activity state.
	//
	function onTick()
	{
		if (didUnload) return;
		if (!didStartActivity) return;

		var now = new Date();
		var needPing = false;
		var timeout = isWindowFocused ? cfg_activityTimeoutFocused : cfg_activityTimeoutUnfocused;
		if (lastActivityTime + timeout > now.getTime())
		{
			// Still active
			inactiveReported = false;

			// Opera needs more frequent pinging because it may leave unexpectedly
			var nextPingTime = lastPingTime * (_window.opera ? 1.5 : 3);
			
			// Don't send a ping more often than every 5 seconds
			if (nextPingTime - lastPingTime < 5000)
			{
				nextPingTime = lastPingTime + 5000;
			}
			
			// Periodically send a signal
			if (getActivityTimespan() >= nextPingTime)
			{
				needPing = true;
				if (cfg_debug >= 1)
				{
					pageData["dbg_lastEvent"] = "onTick active";
					pageData["dbg_lastPingTime"] = lastPingTime;
					pageData["dbg_nextPingTime"] = nextPingTime;
				}
			}
		}
		else if (!inactiveReported)
		{
			setActivity(false);
			
			// When inactive, only send a ping once, then not before being active again
			if (getActivityTimespan() > lastPingTime + 2000)
			{
				// Only send the time if it hasn't already been sent within the last 2 seconds
				needPing = true;
			}
			inactiveReported = true;
			if (cfg_debug >= 1) pageData["dbg_lastEvent"] = "onTick inactive";
		}
		
		if (needPing)
		{
			postData();
		}
	}
	
	// Event handler for click events on link elements. Also called when pressing the Enter key on
	// a link element.
	//
	function onLinkClicked(event)
	{
		//if (cfg_debug >= 1) console.log("onLinkClicked, e.type=" + event.type);

		// Only react on left or middle mouse button clicks. This is far from being perfect at
		// recognising real navigation inputs, but there's not much we can do in a browser.
		if (event.button != 0 && event.button != 1) return;

		// Find the actual link element up the DOM tree
		var element = event.target;
		while (element && !element.tagName.match(/^a(rea)?$/i))
		{
			element = element.parentNode;
		}
		if (!element) return;   // Something went wrong, we're lost.

		var url = element.href;
		if (url && !url.match(/^((java|j|vb)script|mailto):/i))
		{
			var now = (new Date()).getTime();
			if (!outlinkTimes[url] ||
				outlinkTimes[url] + 2000 < now)
			{
				outlinkTimes[url] = now;

				var hostName = (element.hostname || extractHostName(url, true)).toLowerCase();
				var section;
				if (element.getAttribute("data-tomo-imgzoom"))
				{
					pageData[key_imgZoomUrl] = url;
					postData();
				}
				else if (section = element.getAttribute("data-tomo-sectionexpand"))
				{
					pageData[key_sectionExpand] = section;
					postData();
				}
				else if (hostName != myDomain)
				{
					// This looks like an outlink to another domain, report the appropriate action
					pageData[key_outlinkUrl] = url;
					lastLeavingActionTime = now;
					postData(true);
				}
				else if (url.match(/\.(7z|aac|as[fx]|avi|bin|csv|dat|deb|dmg|doc[mx]?|dot|exe|gif|gz|g?zip|hqx|jar|jpe?g|js|mp(2|3|4|e?g)|mov|ms[ip]|od[bfgpst]|og[gv]|pdf|png|pp[st]x?|psd?|qtm?|ra[mr]?|rpm|tar|t?bz2?|tgz|tiff?|torrent|txt|wav|wm[av]|xls[mx]?|xlt|xml|z)$/i))
				{
					// This looks like a download link, report the appropriate action
					pageData[key_downloadUrl] = url;
					lastLeavingActionTime = now;
					postData(true);
				}
			}
		}
	}
	

	// =======================================  PUBLIC API  =======================================


	var _tomo = {

		// Automatically calls the start function upon loading if configured.
		//
		// If this script file was loaded appending the "autostart=1" parameter, the tracking is
		// automatically started. Otherwise the function start needs to be called to start the
		// tracking. This feature is particularly useful to minimise the code inserted on the web
		// page if no further tracking configuration is required.
		//
		autostart: function()
		{
			if (_document.getElementById(cfg_scriptId).src.match(/[?&]autostart=1(&|$)/))
			{
				_tomo.start();
			}
		},
		
		// Starts the tracking process.
		//
		// Before calling this function, all runtime configuration calls to this API must be done
		// so that it can take effect on what this function starts off.
		//
		start: function()
		{
			init();
		},
		
		setPageLanguage: function(lang)
		{
			// Only save data for sending but don't trigger sending. This function must be called
			// before start() or the language data may be sent later or not at all.
			// NOTE: If the display language has already been detected from a translation service
			//       website, this function call is ignored as it is probably wrong.
			if (!isSet(pageLanguage))
			{
				pageLanguage = lang;
				pageData[key_pageLanguage] = lang;
			}
		},
		
		setVisitorVariable: function(name, value)
		{
			// Queue data for sending, register timeout for sending if no ping is sent in time
			pageData["vv_" + name] = value;
			scheduleCVSend();
		},

		setSessionVariable: function(name, value)
		{
			// Queue data for sending, register timeout for sending if no ping is sent in time
			pageData["sv_" + name] = value;
			scheduleCVSend();
		},

		setPageviewVariable: function(name, value)
		{
			// Queue data for sending, register timeout for sending if no ping is sent in time
			pageData["pv_" + name] = value;
			scheduleCVSend();
		},
		
		ignoreLinkClass: function(className)
		{
			ignoredLinkClasses.push(className);
		},
		
		imgZoom: function(url)
		{
			pageData[key_imgZoomUrl] = url;
			postData();
		},
		
		sectionExpand: function(section)
		{
			pageData[key_sectionExpand] = section;
			postData();
		},
		
		siteSearch: function(search)
		{
			pageData[key_siteSearch] = search;
			postData(true);
		},
		
		action: function(type, url, data, leaving)
		{
			if (isSet(url))
				pageData["ac_" + type + "_u"] = url;
			if (isSet(data))
				pageData["ac_" + type + "_d"] = data;
			if (!isSet(url) && !isSet(data))
				pageData["ac_" + type + "_d"] = "";
			if (leaving)
			{
				lastLeavingActionTime = (new Date()).getTime();
				postData(true);
			}
			else
			{
				postData();
			}
		},
		
		newPage: function(newPageviewVars)
		{
			// Notify the server that the page has replaced significant parts of its content and the reported
			// pageview shall be duplicated with new pageview variables
			
			pageData[key_duplicatePageview] = 1;
			pageData["n_" + key_pageTitle] = _document.title;
			// Page URL may have changed, always include it
			pageData["n_" + key_pageUrl] = _location.href;
			
			// Set new pageview variables
			if (isSet(newPageviewVars))
			{
				for (var property in newPageviewVars)
				{
					pageData["n_pv_" + property] = newPageviewVars[property];
				}
			}
			
			postData();

			// Reset variables to start a new pageview
			resetVars();

			// When (interactively) changing the page, assume that the user is currently active
			setActivity(true);
		}

	}; // _tomo = tòmò's object
	

	// Check if autostart is requested
	_tomo.autostart();


	return _tomo;
}();
